# -*- coding: utf-8 -*-
"""scheduler module methods."""
from apscheduler.job import Job as APJob
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

from pytz import utc


def construct_trigger(trigger_type, bg_trigger):
    """Construct an apscheduler trigger based on type and mongo document."""
    trigger_type = trigger_type
    trigger_kwargs = bg_trigger.get_scheduler_kwargs()
    if trigger_type == 'date':
        return DateTrigger(**trigger_kwargs)
    elif trigger_type == 'interval':
        return IntervalTrigger(**trigger_kwargs)
    elif trigger_type == 'cron':
        return CronTrigger(**trigger_kwargs)
    else:
        raise ValueError('Invalid trigger type %s' % trigger_type)


def db_to_scheduler(document, scheduler, alias='beer_garden'):
    """Convert a database job to a scheduler's job."""
    job = APJob.__new__(APJob)
    if document.next_run_time:
        next_run_time = utc.localize(document.next_run_time)
    else:
        next_run_time = None
    state = {
        'id': document.id,
        'func': 'brew_view.scheduler.runner:run_job',
        'trigger': construct_trigger(document.trigger_type, document.trigger),
        'executor': 'default',
        'args': (),
        'kwargs': {
            'request_template': document.request_template,
            'job_id': str(document.id),
        },
        'name': document.name,
        'misfire_grace_time': document.misfire_grace_time,
        'coalesce': document.coalesce,
        'max_instances': 3,
        'next_run_time': next_run_time,
    }
    job.__setstate__(state)
    job._scheduler = scheduler
    job._jobstore_alias = alias
    return job



import datetime
import logging

import brew_view
from bg_utils.models import Request, Job
from bg_utils.parser import BeerGardenSchemaParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError
from brewtils.models import Events, Request as BrewtilsRequest


class RequestAPI(BaseHandler):

    parser = BeerGardenSchemaParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.REQUEST_READ])
    def get(self, request_id):
        """
        ---
        summary: Retrieve a specific Request
        parameters:
          - name: request_id
            in: path
            required: true
            description: The ID of the Request
            type: string
        responses:
          200:
            description: Request with the given ID
            schema:
              $ref: '#/definitions/Request'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.logger.debug("Getting Request: %s", request_id)

        req = Request.objects.get(id=str(request_id))
        req.children = Request.objects(parent=req)

        self.write(self.parser.serialize_request(req, to_string=False))

    @authenticated(permissions=[Permissions.REQUEST_UPDATE])
    def patch(self, request_id):
        """
        ---
        summary: Partially update a Request
        description: |
          The body of the request needs to contain a set of instructions detailing the updates to
          apply. Currently the only operation supported is `replace`, with paths `/status`,
          `/output`, and `/error_class`:
          ```JSON
          {
            "operations": [
              { "operation": "replace", "path": "/status", "value": "" }
            ]
          }
          ```
        parameters:
          - name: request_id
            in: path
            required: true
            description: The ID of the Request
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the Request
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Request with the given ID
            schema:
              $ref: '#/definitions/Request'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        req = Request.objects.get(id=request_id)
        operations = self.parser.parse_patch(self.request.decoded_body, many=True, from_string=True)
        wait_condition = None

        # We note the status before the operations, because it is possible for the operations to
        # update the status of the request. In that case, because the updates are coming in in a
        # single request it is okay to update the output or error_class. Ideally this would be
        # handled correctly when we better integrate PatchOperations with their models.
        status_before = req.status

        for op in operations:
            if op.operation == 'replace':
                if op.path == '/status':
                    if op.value.upper() in BrewtilsRequest.STATUS_LIST:
                        req.status = op.value.upper()

                        if op.value.upper() == 'IN_PROGRESS':
                            self.request.event.name = Events.REQUEST_STARTED.name

                        elif op.value.upper() in BrewtilsRequest.COMPLETED_STATUSES:
                            self.request.event.name = Events.REQUEST_COMPLETED.name

                            if request_id in brew_view.request_map:
                                wait_condition = brew_view.request_map[request_id]
                    else:
                        error_msg = "Unsupported status value '%s'" % op.value
                        self.logger.warning(error_msg)
                        raise ModelValidationError(error_msg)
                elif op.path == '/output':
                    if req.output == op.value:
                        continue

                    if status_before in Request.COMPLETED_STATUSES:
                        raise ModelValidationError("Cannot update output for a request "
                                                   "that is already completed")
                    req.output = op.value
                elif op.path == '/error_class':
                    if req.error_class == op.value:
                        continue

                    if status_before in Request.COMPLETED_STATUSES:
                        raise ModelValidationError("Cannot update error_class for a "
                                                   "request that is already completed")
                    req.error_class = op.value
                    self.request.event.error = True
                else:
                    error_msg = "Unsupported path '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError(error_msg)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError(error_msg)

        req.save()
        self._update_metrics(req, status_before)
        self._update_job_numbers(req, status_before)

        if wait_condition:
            wait_condition.notify()

        self.request.event_extras = {'request': req, 'patch': operations}

        self.write(self.parser.serialize_request(req, to_string=False))

    def _update_job_numbers(self, request, status_before):
        if (
            not request.metadata.get('_bg_job_id') or
            status_before == request.status or
            request.status not in Request.COMPLETED_STATUSES
        ):
            return

        try:
            job_id = request.metadata.get('_bg_job_id')
            document = Job.objects.get(id=job_id)
            if request.status == 'ERROR':
                document.error_count += 1
            elif request.status == 'SUCCESS':
                document.success_count += 1
            document.save()
        except Exception as exc:
            self.logger.warning('Could not update job counts.')
            self.logger.exception(exc)

    def _update_metrics(self, request, status_before):
        """Update metrics associated with this new request.

        We have already confirmed that this request has had a valid state
        transition because updating the status verifies that the updates are
        correct.

        In addition, this call is happening after the save to the
        database. Now we just need to update the metrics.
        """
        if (
            status_before == request.status or
            status_before in BrewtilsRequest.COMPLETED_STATUSES
        ):
            return

        labels = {
            'system': request.system,
            'system_version': request.system_version,
            'instance_name': request.instance_name,
        }

        if status_before == 'CREATED':
            self.queued_request_gauge.labels(**labels).dec()
        elif status_before == 'IN_PROGRESS':
            self.in_progress_request_gauge.labels(**labels).dec()

        if request.status == 'IN_PROGRESS':
            self.in_progress_request_gauge.labels(**labels).inc()

        elif request.status in BrewtilsRequest.COMPLETED_STATUSES:
            # We don't use _measure_latency here because the request times are
            # stored in UTC and we need to make sure we're comparing apples to
            # apples.
            latency = (datetime.datetime.utcnow() - request.created_at).total_seconds()
            labels['command'] = request.command
            labels['status'] = request.status

            self.completed_request_counter.labels(**labels).inc()
            self.plugin_command_latency.labels(**labels).observe(latency)



import logging

from tornado.gen import coroutine

import brew_view
from brew_view import thrift_context
from brew_view.base_handler import BaseHandler


class ConfigHandler(BaseHandler):

    def get(self):
        """Subset of configuration options that the frontend needs"""
        configs = {
            'allow_unsafe_templates': brew_view.config.application.allow_unsafe_templates,
            'application_name': brew_view.config.application.name,
            'amq_admin_port': brew_view.config.amq.connections.admin.port,
            'amq_host': brew_view.config.amq.host,
            'amq_port': brew_view.config.amq.connections.message.port,
            'amq_virtual_host': brew_view.config.amq.virtual_host,
            'backend_host': brew_view.config.backend.host,
            'backend_port': brew_view.config.backend.port,
            'icon_default': brew_view.config.application.icon_default,
            'debug_mode': brew_view.config.debug_mode,
            'url_prefix': brew_view.config.web.url_prefix,
            'metrics_url': brew_view.config.metrics.url,
            'auth_enabled': brew_view.config.auth.enabled,
        }
        self.write(configs)


class VersionHandler(BaseHandler):

    @coroutine
    def get(self):
        with thrift_context() as client:
            try:
                bartender_version = yield client.getVersion()
            except Exception as ex:
                logger = logging.getLogger(__name__)
                logger.error("Could not get Bartender Version.")
                logger.exception(ex)
                bartender_version = "unknown"

        self.write({
            "brew_view_version": brew_view.__version__,
            "bartender_version": bartender_version,
            "current_api_version": "v1",
            "supported_api_versions": ["v1"]
        })


class SwaggerConfigHandler(BaseHandler):

    def get(self):
        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.write({
            'url': brew_view.config.web.url_prefix + 'api/v1/spec',
            'validatorUrl': None
        })


class SpecHandler(BaseHandler):

    def get(self):
        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.write(brew_view.api_spec.to_dict())



import re
import socket

import time
from jwt.exceptions import ExpiredSignatureError, InvalidSignatureError
from mongoengine.errors import (
    DoesNotExist, NotUniqueError, ValidationError as MongoValidationError)
from prometheus_client import Gauge, Counter, Summary
from thriftpy.thrift import TException
from tornado.web import HTTPError, RequestHandler

import bg_utils
import brew_view
from brew_view.authorization import AuthMixin
from brewtils.errors import (
    ModelError, ModelValidationError, RequestForbidden, RequestPublishException,
    WaitExceededError, AuthorizationRequired)
from brewtils.models import Event


class BaseHandler(AuthMixin, RequestHandler):
    """Base handler from which all handlers inherit"""

    MONGO_ID_PATTERN = r'.*/([0-9a-f]{24}).*'

    charset_re = re.compile(r'charset=(.*)$')

    error_map = {
        MongoValidationError: {'status_code': 400},
        ModelError: {'status_code': 400},
        bg_utils.bg_thrift.InvalidSystem: {'status_code': 400},
        ExpiredSignatureError: {'status_code': 401},
        AuthorizationRequired: {'status_code': 401},
        RequestForbidden: {'status_code': 403},
        InvalidSignatureError: {'status_code': 403},
        DoesNotExist: {'status_code': 404, 'message': 'Resource does not exist'},
        WaitExceededError: {'status_code': 408, 'message': 'Max wait time exceeded'},
        NotUniqueError: {'status_code': 409, 'message': 'Resource already exists'},

        RequestPublishException: {'status_code': 502},
        bg_utils.bg_thrift.BaseException: {'status_code': 502, 'message': 'An error occurred '
                                                                          'on the backend'},
        TException: {'status_code': 503, 'message': 'Could not connect to Bartender'},
        socket.timeout: {'status_code': 504, 'message': 'Backend request timed out'},
    }

    # Prometheus metrics.
    # Summaries:
    http_api_latency_total = Summary(
        'bg_http_api_latency_seconds',
        'Total number of seconds each API endpoint is taking to respond.',
        ['method', 'route', 'status']
    )
    plugin_command_latency = Summary(
        'bg_plugin_command_latency_seconds',
        'Total time taken for a command to complete in seconds.',
        ['system', 'instance_name', 'system_version', 'command', 'status'],
    )

    # Counters:
    completed_request_counter = Counter(
        'bg_completed_requests_total',
        'Number of completed requests.',
        ['system', 'instance_name', 'system_version', 'command', 'status'],
    )
    request_counter_total = Counter(
        'bg_requests_total',
        'Number of requests.',
        ['system', 'instance_name', 'system_version', 'command'],
    )

    # Gauges:
    queued_request_gauge = Gauge(
        'bg_queued_requests',
        'Number of requests waiting to be processed.',
        ['system', 'instance_name', 'system_version'],
    )
    in_progress_request_gauge = Gauge(
        'bg_in_progress_requests',
        'Number of requests IN_PROGRESS',
        ['system', 'instance_name', 'system_version'],
    )

    def set_default_headers(self):
        """Headers set here will be applied to all responses"""
        self.set_header("BG-Version", brew_view.__version__)

        if brew_view.config.cors_enabled:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Headers", "Content-Type")
            self.set_header("Access-Control-Allow-Methods", "GET, POST, PATCH, DELETE, OPTIONS")

    @property
    def prometheus_endpoint(self):
        """Removes Mongo ID from endpoint."""
        to_return = self.request.path.rstrip('/')
        for mongo_id in re.findall(self.MONGO_ID_PATTERN, self.request.path):
            to_return = to_return.replace(mongo_id, '<ID>')
        return to_return

    def prepare(self):
        """Called before each verb handler"""

        # Used for recording prometheus metrics.
        # We keep this time in seconds, also time-zone does not matter
        # because we are just calculating a duration.
        self.request.created_time = time.time()

        # This is used for sending event notifications
        self.request.event = Event()
        self.request.event_extras = {}

        content_type = self.request.headers.get('content-type', '')
        if self.request.method.upper() in ['POST', 'PATCH'] and content_type:
            content_type = content_type.split(';')

            self.request.mime_type = content_type[0]
            if self.request.mime_type not in ['application/json',
                                              'application/x-www-form-urlencoded']:
                raise ModelValidationError('Unsupported or missing content-type header')

            # Attempt to parse out the charset and decode the body, default to utf-8
            charset = 'utf-8'
            if len(content_type) > 1:
                search_result = self.charset_re.search(content_type[1])
                if search_result:
                    charset = search_result.group(1)
            self.request.charset = charset
            self.request.decoded_body = self.request.body.decode(charset)

    def on_finish(self):
        """Called after a handler completes processing"""
        self.http_api_latency_total.labels(
            method=self.request.method.upper(),
            route=self.prometheus_endpoint,
            status=self.get_status(),
        ).observe(self._measure_latency())

        if self.request.event.name:
            brew_view.event_publishers.publish_event(self.request.event,
                                                     **self.request.event_extras)

    def _measure_latency(self):
        """We measure the latency in seconds as a float."""
        return time.time() - self.request.created_time

    def options(self, *args, **kwargs):

        if brew_view.config.cors_enabled:
            self.set_status(204)
        else:
            raise HTTPError(403, reason='CORS is disabled')

    def write_error(self, status_code, **kwargs):
        """Transform an exception into a response.

        This protects controllers from having to write a lot of the same code over and over and
        over. Controllers can, of course, overwrite error handlers and return their own responses
        if necessary, but generally, this is where error handling should occur.

        When an exception is handled this function makes two passes through error_map. The first
        pass is to see if the exception type can be matched exactly. If there is no exact type
        match the second pass will attempt to match using isinstance. If a message is provided in
        the error_map it takes precedence over the exception message.

        ***NOTE*** Nontrivial inheritance trees will almost definitely break. This is a BEST EFFORT
        using a simple isinstance check on an unordered data structure. So if an exception class
        has both a parent and a grandparent in the error_map there is no guarantee about which
        message / status code will be chosen. The same applies to exceptions that use multiple
        inheritance.

        ***LOGGING***
        An exception raised in a controller method will generate logging to the tornado.application
        logger that includes a stacktrace. That logging occurs before this method is invoked.
        The result of this method will generate logging to the tornado.access logger as usual.
        So there is no need to do additional logging here as the 'real' exception will already have
        been logged.

        :param status_code: a status_code that will be used if no match is found in the error map
        :return: None
        """
        code = 0
        message = ''

        if 'exc_info' in kwargs:
            typ3 = kwargs['exc_info'][0]
            e = kwargs['exc_info'][1]

            error_dict = None
            if typ3 in self.error_map.keys():
                error_dict = self.error_map[typ3]
            else:
                for error_type in self.error_map.keys():
                    if isinstance(e, error_type):
                        error_dict = self.error_map[error_type]
                        break

            if error_dict:
                code = error_dict.get('status_code', 500)
                message = error_dict.get('message', str(e))

            elif brew_view.config.debug_mode:
                message = str(e)

        code = code or status_code or 500
        message = message or ('Encountered unknown exception. Please check '
                              'with your System Administrator.')

        self.request.event.error = True
        self.request.event.payload = {'message': message}

        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.set_status(code)
        self.finish({'message': message})



import logging
from datetime import timedelta
from functools import wraps

from pika import BasicProperties
from pika import TornadoConnection
from six import string_types
from tornado.concurrent import return_future
from tornado.gen import coroutine
from tornado.httpclient import AsyncHTTPClient
from tornado.ioloop import IOLoop
from tornado.queues import Queue

import brew_view
from bg_utils.event_publisher import EventPublisher
from bg_utils.parser import BeerGardenSchemaParser
from bg_utils.pika import get_routing_key
from brewtils.models import Events
from brewtils.queues import PikaClient
from brewtils.rest.client import RestClient


class BeergardenPublisher(EventPublisher):

    def __init__(self):
        self._client = RestClient(brew_view.config.web.public_fqdn,
                                  brew_view.config.web.port,
                                  ssl_enabled=brew_view.config.web.ssl.enabled,
                                  url_prefix=brew_view.config.web.url_prefix)

    def _event_prepare(self, event, **kwargs):

        if event.name.startswith('REQUEST'):
            request = kwargs.pop('request', None)
            if request:
                event.metadata['entity_url'] = self._client.request_url + str(request.id)
                event.payload = {k: str(getattr(request, k)) for k in
                                 ['id', 'command', 'system', 'system_version', 'instance_name']}

        elif event.name.startswith('SYSTEM'):
            system = kwargs.pop('system', None)
            if system:
                event.metadata['entity_url'] = self._client.system_url + str(system.id)
                event.payload = {'id': str(system.id)}

        elif event.name.startswith('INSTANCE'):
            instance = kwargs.pop('instance', None)
            if instance:
                event.metadata['entity_url'] = self._client.instance_url + str(instance.id)
                event.payload = {'id': str(instance.id)}

        return event


class HttpPublisher(EventPublisher):

    def __init__(self, urls, ssl_context=None):
        self._client = AsyncHTTPClient(
            defaults={'ssl_options': ssl_context} if ssl_context else None)
        self._urls = urls

    def publish(self, message, **kwargs):
        for url in self._urls:
            self._client.fetch(url, raise_error=False, method='POST',
                               headers={'content-type': 'application/json'},
                               body=message)


class RequestPublisher(EventPublisher):

    def __init__(self, ssl_context=None):
        self._client = AsyncHTTPClient(
            defaults={'ssl_options': ssl_context} if ssl_context else None)

    def publish(self, message, **kwargs):
        for url in kwargs.get('urls', []):
            self._client.fetch(url, raise_error=False, method='POST',
                               headers={'content-type': 'application/json'},
                               body=message)

    def _event_publish_args(self, event, **kwargs):
        urls = []

        if event.name.startswith('REQUEST') and 'request' in kwargs:
            if 'webhooks' in kwargs['request'].metadata:
                webhook_dict = kwargs['request'].metadata['webhooks']

                # Always add current event urls
                urls += webhook_dict.get(event.name, [])

                # Additionally do some quick translation for success / failure completion events
                if event.name == Events.REQUEST_COMPLETED.name:
                    urls += webhook_dict.get(
                        'REQUEST_FAILURE' if event.error else 'REQUEST_SUCCESS', [])

        return {'urls': urls} if urls else {}


class WebsocketPublisher(BeergardenPublisher):
    """Publisher implementation that publishes to a websocket"""

    def __init__(self, socket_class):
        BeergardenPublisher.__init__(self)

        self._socket = socket_class

    def publish(self, message, **kwargs):
        self._socket.publish(message)

    def shutdown(self):
        self._socket.shutdown()


class MongoPublisher(BeergardenPublisher):
    """Publisher implementation that 'publishes' to Mongo"""

    def publish(self, message, **kwargs):
        message.save()

    def _event_serialize(self, event, **kwargs):
        return BeerGardenSchemaParser.parse_event(
            BeerGardenSchemaParser.serialize_event(event, to_string=False)
        )


class TornadoPikaPublisher(BeergardenPublisher, PikaClient):

    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)

        self._shutdown_timeout = timedelta(seconds=kwargs.pop('shutdown_timeout', 5))
        self._work_queue = Queue()
        self._connection = None
        self._channel = None

        self.coroutiner = CoroutineMaker({
            'TornadoConnection': 'on_open_callback',
            'channel': 'on_open_callback'
        })

        # Trying to get super() to work with incompatible signatures is a nightmare
        BeergardenPublisher.__init__(self)
        PikaClient.__init__(self, **kwargs)

        IOLoop.current().spawn_callback(self._process)

    def shutdown(self):
        return self._work_queue.join(timeout=self._shutdown_timeout)

    @coroutine
    def _open_connection(self):
        self._connection = yield self.coroutiner.convert(TornadoConnection)(
            parameters=self._conn_params,
            stop_ioloop_on_close=False)

    @coroutine
    def _open_channel(self):
        self._channel = yield self.coroutiner.convert(self._connection.channel)()

    @coroutine
    def _process(self):

        while True:
            item = yield self._work_queue.get()

            try:
                if not self._connection or not self._connection.is_open:
                    yield self._open_connection()
                if not self._channel or not self._channel.is_open:
                    yield self._open_channel()

                yield getattr(self._channel, item[0])(**item[1])
            finally:
                self._work_queue.task_done()

    def publish(self, message, **kwargs):
        """Publish a message.

        :param message: The message to publish
        :param kwargs: Additional message properties
        :Keyword Arguments:
            * *routing_key* --
              Routing key to use when publishing
            * *headers* --
              Headers to be included as part of the message properties
            * *expiration* --
              Expiration to be included as part of the message properties
        :return: None
        """
        self._work_queue.put(('basic_publish', {
            'exchange': self._exchange, 'routing_key': kwargs['routing_key'], 'body': message,
            'properties': BasicProperties(app_id='beer-garden', content_type='text/plain',
                                          headers=kwargs.pop('headers', None),
                                          expiration=kwargs.pop('expiration', None))
        }))

    def _event_publish_args(self, event, **kwargs):

        # Main thing we need to do here is figure out the appropriate routing key
        args = {}
        if event.metadata and 'routing_key' in event.metadata:
            args['routing_key'] = event.metadata['routing_key']
        elif 'request' in kwargs:
            request = kwargs['request']
            args['routing_key'] = get_routing_key('request', request.system, request.system_version,
                                                  request.instance_name)
        else:
            args['routing_key'] = 'beergarden'

        return args


class CoroutineMaker(object):
    """Helper class to wrap functions in Tornado futures

    Tornado has a nice @return_future decorator that converts a callback-based function into a
    Tornado async one. The issue is that it expects the wrapped function to take a `callback`
    argument and invoke that argument when execution is completed.

    We need to convert a couple of pika functions in this way. Unfortunately, they're expecting a
    callback function to be provided with the `on_open_callback` keyword parameter. So we use this
    class to wrap the functions and mangle the arguments when the function is called to pass the
    Tornado-provided callback to the function like it's expecting.
    """

    def __init__(self, config):
        self._signatures = {}

        for func, param in config.items():
            if isinstance(param, int):
                self._signatures[func] = self._positional_signature(param)

            elif isinstance(param, string_types):
                self._signatures[func] = self._keyword_signature(param)

            else:
                raise Exception("Bad configuration")

    def convert(self, func):

        @return_future
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                args, kwargs = self._signatures[func.__name__](args, kwargs)
            except KeyError:
                pass
            func(*args, **kwargs)

        return wrapper

    @staticmethod
    def _positional_signature(position):

        def signature(args, kwargs):
            args = list(args)
            args.insert(position, kwargs.pop('callback'))
            return args, kwargs

        return signature

    @staticmethod
    def _keyword_signature(keyword):

        def signature(args, kwargs):
            kwargs[keyword] = kwargs.pop('callback')
            return args, kwargs

        return signature



import logging

import mongoengine.errors

from bg_utils.models import Command
from bg_utils.parser import BeerGardenSchemaParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler


class CommandListAPI(BaseHandler):

    parser = BeerGardenSchemaParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.COMMAND_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Commands
        responses:
          200:
            description: All Commands
            schema:
              type: array
              items:
                $ref: '#/definitions/Command'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Commands
        """
        self.logger.debug("Getting Commands")

        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        try:
            self.write(self.parser.serialize_command(Command.objects.all(), many=True,
                                                     to_string=True))
        except mongoengine.errors.DoesNotExist as ex:
            self.logger.error("Got an error while attempting to serialize commands. "
                              "This error usually indicates "
                              "there are orphans in the database.")
            raise mongoengine.errors.InvalidDocumentError(ex)



import json
import logging
from datetime import timedelta
from functools import reduce

from mongoengine import Q
from tornado.gen import coroutine
from tornado.locks import Condition

import bg_utils
import brew_view
from bg_utils.models import Request, System
from bg_utils.parser import BeerGardenSchemaParser
from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import (ModelValidationError, RequestPublishException,
                             TimeoutExceededError)
from brewtils.models import Events


class RequestListAPI(BaseHandler):

    parser = BeerGardenSchemaParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.REQUEST_READ])
    def get(self):
        """
        ---
        summary: Retrieve a page of all Requests
        description: |
          This endpoint queries multiple requests at once. Because it's intended to be used with
          Datatables the query parameters are ... complicated. Here are things to keep in mind:

          * With no query parameters this endpoint will return the first 100 non-child requests.
            This can be controlled by passing the `start` and `length` query parameters.

          * This endpoint does NOT return child request definitions. If you want to see child
            requests you must use the /api/v1/requests/{request_id} endpoint.

          * By default this endpoint also does not include child requests in the response. That
            is, if a request has a non-null `parent` field it will not be included in the response
            array. Use the `include_children` query parameter to change this behavior.

          To filter, search, and order you need to conform to how Datatables structures its query
          parameters.

          * To indicate fields that should be included in the response specify multiple `columns`
            query parameters:
          ```JSON
          {
            "data": "command",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value":"","regex":false}
          }
          {
            "data": "system",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value": "","regex": false}
          }
          ```
          * To filter a specific field set the value in the `search` key of its `column` definition:
          ```JSON
          {
            "data": "status",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value": "SUCCESS", "regex":false}
          }
          ```

          * To sort by a field use the `order` parameter. The `column` value should be the index
            of the column to sort and the `dir` value can be either "asc" or "desc."
          ```JSON
          {"column": 3,"dir": "asc"}
          ```

          * To perform a text-based search across all fields use the `search` parameter:
          ```JSON
          { "value": "SEARCH VALUE", "regex": false }
          ```
        parameters:
          - name: include_children
            in: query
            required: false
            description: Flag indicating whether to include child requests in the response list
            type: boolean
            default: false
          - name: start
            in: query
            required: false
            description: The starting index for the page
            type: integer
          - name: length
            in: query
            required: false
            description: The maximum number of Requests to include in the page
            type: integer
          - name: draw
            in: query
            required: false
            description: Used by datatables, will be echoed in a response header
            type: integer
          - name: columns
            in: query
            required: false
            description: Datatables column definitions
            type: array
            collectionFormat: multi
            items:
              properties:
                data:
                  type: string
                name:
                  type: string
                searchable:
                  type: boolean
                  default: true
                orderable:
                  type: boolean
                  default: true
                search:
                  properties:
                    value:
                      type: string
                    regex:
                      type: boolean
                      default: false
          - name: search
            in: query
            required: false
            description: Datatables search object
            type: string
          - name: order
            in: query
            required: false
            description: Datatables order object
            type: string
        responses:
          200:
            description: A page of Requests
            schema:
              type: array
              items:
                $ref: '#/definitions/Request'
            headers:
              start:
                type: integer
                description: Echo of 'start' query parameter or '0'
              length:
                type: integer
                description: Number of Requests in the response
              draw:
                type: integer
                description: Echo of the 'draw' query parameter
              recordsFiltered:
                type: integer
                description: The number of Requests that satisfied the search filters
              recordsTotal:
                type: integer
                description: The total number of Requests
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.logger.debug("Getting Requests")

        # We need to do the slicing on the query. This greatly reduces load time.
        start = int(self.get_argument('start', default=0))
        length = int(self.get_argument('length', default=100))
        requests, filtered_count, requested_fields = self._get_requests(start, start+length)

        # Sweet, we have data. Now setup some headers for the response
        response_headers = {
            # These are a courtesy for non-datatables requests. We want people making a request
            # with no headers to realize they probably aren't getting the full dataset
            'start': start,
            'length': len(requests),

            # And these are required by datatables
            'recordsFiltered': filtered_count,
            'recordsTotal': Request.objects.count(),
            'draw': self.get_argument('draw', '')
        }

        for key, value in response_headers.items():
            self.add_header(key, value)
            self.add_header('Access-Control-Expose-Headers', key)

        self.set_header('Content-Type', 'application/json; charset=UTF-8')
        self.write(self.parser.serialize_request(requests, to_string=True, many=True,
                                                 only=requested_fields))

    @coroutine
    @authenticated(permissions=[Permissions.REQUEST_CREATE])
    def post(self):
        """
        ---
        summary: Create a new Request
        parameters:
          - name: request
            in: body
            description: The Request definition
            schema:
              $ref: '#/definitions/Request'
          - name: blocking
            in: query
            required: false
            description: Flag indicating whether to wait for request completion
            type: boolean
            default: false
          - name: timeout
            in: query
            required: false
            description: Maximum time (seconds) to wait for request completion
            type: integer
            default: None (Wait forever)
        consumes:
          - application/json
          - application/x-www-form-urlencoded
        responses:
          201:
            description: A new Request has been created
            schema:
              $ref: '#/definitions/Request'
            headers:
              Instance-Status:
                type: string
                description: Current status of the Instance that will process the created Request
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.request.event.name = Events.REQUEST_CREATED.name

        if self.request.mime_type == 'application/json':
            request_model = self.parser.parse_request(self.request.decoded_body, from_string=True)
        elif self.request.mime_type == 'application/x-www-form-urlencoded':
            args = {'parameters': {}}
            for key, value in self.request.body_arguments.items():
                if key.startswith('parameters.'):
                    args['parameters'][key.replace('parameters.', '')] = value[0].decode(
                        self.request.charset)
                else:
                    args[key] = value[0].decode(self.request.charset)
            request_model = Request(**args)
        else:
            raise ModelValidationError('Unsupported or missing content-type header')

        if request_model.parent:
            request_model.parent = Request.objects.get(id=str(request_model.parent.id))
            request_model.has_parent = True
        else:
            request_model.has_parent = False

        if self.current_user:
            request_model.requester = self.current_user.username

        with thrift_context() as client:
            try:
                request_model.save()
                yield client.processRequest(str(request_model.id))

            except bg_utils.bg_thrift.InvalidRequest as ex:
                request_model.delete()
                raise ModelValidationError(ex.message)
            except bg_utils.bg_thrift.PublishException as ex:
                request_model.delete()
                raise RequestPublishException(ex.message)
            except Exception:
                if request_model.id:
                    request_model.delete()
                raise
            else:
                if self.get_argument('blocking', default='').lower() == 'true':
                    timeout = self.get_argument('timeout', default=None)
                    if timeout is None:
                        timeout_delta = None
                    else:
                        timeout_delta = timedelta(seconds=int(timeout))

                    condition = Condition()
                    brew_view.request_map[str(request_model.id)] = condition

                    wait_result = yield condition.wait(timeout_delta)
                    if not wait_result:
                        raise TimeoutExceededError()

        request_model.reload()

        # Query for request from body id
        req = Request.objects.get(id=str(request_model.id))

        # Now attempt to add the instance status as a header.
        # The Request is already created at this point so it's a best-effort thing
        self.set_header("Instance-Status", 'UNKNOWN')

        try:
            # Since request has system info we can query for a system object
            system = System.objects.get(name=req.system, version=req.system_version)

            # Loop through all instances in the system until we find the instance that matches
            # the request instance
            for instance in system.instances:
                if instance.name == req.instance_name:
                    self.set_header("Instance-Status", instance.status)

        # The Request is already created at this point so adding the Instance status header is a
        # best-effort thing
        except Exception as ex:
            self.logger.exception('Unable to get Instance status for Request %s: %s',
                                  str(request_model.id), ex)

        self.request.event_extras = {'request': req}

        self.set_status(201)
        self.queued_request_gauge.labels(
            system=request_model.system,
            system_version=request_model.system_version,
            instance_name=request_model.instance_name,
        ).inc()
        self.request_counter_total.labels(
            system=request_model.system,
            system_version=request_model.system_version,
            instance_name=request_model.instance_name,
            command=request_model.command,
        ).inc()
        self.write(self.parser.serialize_request(request_model, to_string=False))

    def _get_requests(self, start, end):
        """Get Requests matching the HTTP request query parameters.

        :return requests: The collection of requests
        :return requested_fields: The fields to be returned for each Request
        """
        search_params = []
        requested_fields = []
        order_by = None
        overall_search = None
        query = Request.objects

        raw_columns = self.get_query_arguments('columns')
        if raw_columns:
            columns = []

            for raw_column in raw_columns:
                column = json.loads(raw_column)
                columns.append(column)

                if column['data']:
                    requested_fields.append(column['data'])

                if 'searchable' in column and column['searchable'] and column['search']['value']:
                    if column['data'] in ['created_at', 'updated_at']:
                        search_dates = column['search']['value'].split('~')

                        if search_dates[0]:
                            search_params.append(Q(**{column['data']+'__gte': search_dates[0]}))

                        if search_dates[1]:
                            search_params.append(Q(**{column['data']+'__lte': search_dates[1]}))
                    else:
                        search_query = Q(**{column['data']+'__contains': column['search']['value']})
                        search_params.append(search_query)

            raw_order = self.get_query_argument('order', default=None)
            if raw_order:
                order = json.loads(raw_order)
                order_by = columns[order.get('column')]['data']
                if order.get('dir') == 'desc':
                    order_by = '-' + order_by

        raw_search = self.get_query_argument('search', default=None)
        if raw_search:
            search = json.loads(raw_search)
            if search['value']:
                overall_search = '"'+search['value']+'"'

        # Default to only top-level requests
        if self.get_query_argument('include_children', default='false').lower() != 'true':
            search_params.append(Q(has_parent=False))

        # Now we can construct the actual query parameters
        query_params = reduce(lambda x, y: x & y, search_params, Q())

        # Further modify the query itself
        if overall_search:
            query = query.search_text(overall_search)

        if order_by:
            query = query.order_by(order_by)

        # Marshmallow treats [] as 'serialize nothing' which is not what we
        # want, so translate to None
        if requested_fields:
            query = query.only(*requested_fields)
        else:
            requested_fields = None

        # Execute the query / count
        requests = query.filter(query_params)
        filtered_count = requests.count()

        # Only return the correct slice of the QuerySet
        return requests[start:end], filtered_count, requested_fields



import re

from setuptools import setup, find_packages


def find_version(version_file):
    version_line = open(version_file, "rt").read()
    match_object = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_line, re.M)

    if not match_object:
        raise RuntimeError("Unable to find version string in %s" % version_file)

    return match_object.group(1)


with open('README.rst') as readme_file:
    readme = readme_file.read()


setup(
    name='brew-view',
    version=find_version('brew_view/_version.py'),
    description='Beergarden Frontend',
    long_description=readme,
    url='https://beer-garden.io/',
    author='The beer-garden Team',
    author_email=' ',
    license='MIT',
    packages=(find_packages(exclude=['test', 'test.*'])),
    include_package_data=True,
    install_requires=[
        "apispec==0.38.0",
        "apscheduler==3.5.1",
        "bg-utils>=2.4.0",
        "brewtils>=2.4.0",
        "prometheus_client==0.3.1",
        "tornado==5.1",
    ],
    extras_require={
        ':python_version=="2.7"': [
            'futures>=3.1.1',
        ]
    },
    entry_points={
        'console_scripts': [
            'generate_brew_view_config=brew_view.__main__:generate_config',
            'migrate_brew_view_config=brew_view.__main__:migrate_config',
            'generate_brew_view_log_config=brew_view.__main__:generate_logging_config',
            'brew-view=brew_view.__main__:main'
        ]
    }
)
